@@ -6,24 +6,15 @@ import re
 import stat
 import weakref
 from collections import deque
+from collections.abc import Callable, Generator, Iterable, Iterator, Sequence
 from contextlib import suppress
 from gettext import gettext as _
 from operator import attrgetter
 from typing import (
     Any,
-    Callable,
     Deque,
-    Dict,
-    Generator,
-    Iterable,
-    Iterator,
-    List,
     NamedTuple,
     Optional,
-    Sequence,
-    Set,
-    Tuple,
-    Union,
 )
 
 from .borders import Border, Borders
@@ -37,8 +28,10 @@ from .fast_data_types import (
     GLFW_RELEASE,
     add_tab,
     attach_window,
+    buffer_keys_in_window,
     current_focused_os_window_id,
     detach_window,
+    focus_os_window,
     get_boss,
     get_click_interval,
     get_options,
@@ -51,14 +44,16 @@ from .fast_data_types import (
     ring_bell,
     set_active_tab,
     set_active_window,
+    set_redirect_keys_to_overlay,
     swap_tabs,
     sync_os_window_title,
 )
 from .layout.base import Layout
 from .layout.interface import create_layout_object_for, evict_cached_layouts
+from .progress import ProgressState
 from .tab_bar import TabBar, TabBarData
 from .types import ac
-from .typing import EdgeLiteral, SessionTab, SessionType, TypedDict
+from .typing_compat import EdgeLiteral, SessionTab, SessionType, TypedDict
 from .utils import cmdline_for_hold, log_error, platform_window_id, resolved_shell, shlex_split, which
 from .window import CwdRequest, Watchers, Window, WindowDict
 from .window_list import WindowList
@@ -69,7 +64,7 @@ class TabMouseEvent(NamedTuple):
     modifiers: int
     action: int
     at: float
-    tab_idx: Optional[int]
+    tab_idx: int | None
 
 
 class TabDict(TypedDict):
@@ -78,36 +73,36 @@ class TabDict(TypedDict):
     is_active: bool
     title: str
     layout: str
-    layout_state: Dict[str, Any]
-    layout_opts: Dict[str, Any]
-    enabled_layouts: List[str]
-    windows: List[WindowDict]
-    groups: List[Dict[str, Any]]
-    active_window_history: List[int]
+    layout_state: dict[str, Any]
+    layout_opts: dict[str, Any]
+    enabled_layouts: list[str]
+    windows: list[WindowDict]
+    groups: list[dict[str, Any]]
+    active_window_history: list[int]
 
 
 class SpecialWindowInstance(NamedTuple):
-    cmd: Optional[List[str]]
-    stdin: Optional[bytes]
-    override_title: Optional[str]
-    cwd_from: Optional[CwdRequest]
-    cwd: Optional[str]
-    overlay_for: Optional[int]
-    env: Optional[Dict[str, str]]
-    watchers: Optional[Watchers]
+    cmd: list[str] | None
+    stdin: bytes | None
+    override_title: str | None
+    cwd_from: CwdRequest | None
+    cwd: str | None
+    overlay_for: int | None
+    env: dict[str, str] | None
+    watchers: Watchers | None
     overlay_behind: bool
     hold: bool
 
 
 def SpecialWindow(
-    cmd: Optional[List[str]],
-    stdin: Optional[bytes] = None,
-    override_title: Optional[str] = None,
-    cwd_from: Optional[CwdRequest] = None,
-    cwd: Optional[str] = None,
-    overlay_for: Optional[int] = None,
-    env: Optional[Dict[str, str]] = None,
-    watchers: Optional[Watchers] = None,
+    cmd: list[str] | None,
+    stdin: bytes | None = None,
+    override_title: str | None = None,
+    cwd_from: CwdRequest | None = None,
+    cwd: str | None = None,
+    overlay_for: int | None = None,
+    env: dict[str, str] | None = None,
+    watchers: Watchers | None = None,
     overlay_behind: bool = False,
     hold: bool = False,
 ) -> SpecialWindowInstance:
@@ -124,18 +119,22 @@ def add_active_id_to_history(items: Deque[int], item_id: int, maxlen: int = 64)
 
 class Tab:  # {{{
 
-    active_fg: Optional[int] = None
-    active_bg: Optional[int] = None
-    inactive_fg: Optional[int] = None
-    inactive_bg: Optional[int] = None
+    active_fg: int | None = None
+    active_bg: int | None = None
+    inactive_fg: int | None = None
+    inactive_bg: int | None = None
     confirm_close_window_id: int = 0
+    num_of_windows_with_progress: int = 0
+    total_progress: int = 0
+    has_indeterminate_progress: bool = False
+    last_focused_window_with_progress_id: int = 0
 
     def __init__(
         self,
         tab_manager: 'TabManager',
         session_tab: Optional['SessionTab'] = None,
-        special_window: Optional[SpecialWindowInstance] = None,
-        cwd_from: Optional[CwdRequest] = None,
+        special_window: SpecialWindowInstance | None = None,
+        cwd_from: CwdRequest | None = None,
         no_initial_window: bool = False
     ):
         self.tab_manager_ref = weakref.ref(tab_manager)
@@ -148,8 +147,8 @@ class Tab:  # {{{
         self.enabled_layouts = [x.lower() for x in getattr(session_tab, 'enabled_layouts', None) or get_options().enabled_layouts]
         self.borders = Borders(self.os_window_id, self.id)
         self.windows: WindowList = WindowList(self)
-        self._last_used_layout: Optional[str] = None
-        self._current_layout_name: Optional[str] = None
+        self._last_used_layout: str | None = None
+        self._current_layout_name: str | None = None
         self.cwd = self.args.directory
         if no_initial_window:
             self._set_current_layout(self.enabled_layouts[0])
@@ -167,6 +166,29 @@ class Tab:  # {{{
             self._set_current_layout(l0)
             self.startup(session_tab)
 
+    def update_progress(self) -> None:
+        self.num_of_windows_with_progress = 0
+        self.total_progress = 0
+        self.last_focused_window_with_progress_id = 0
+        self.has_indeterminate_progress = False
+        focused_at = 0.
+        for window in self:
+            p = window.progress
+            if p.state is ProgressState.unset:
+                continue
+            if p.state in (ProgressState.set, ProgressState.paused):
+                self.total_progress += p.percent
+                self.num_of_windows_with_progress += 1
+            elif p.state is ProgressState.indeterminate:
+                self.has_indeterminate_progress = True
+            if window.last_focused_at > focused_at or (not window.last_focused_at and window.id > self.last_focused_window_with_progress_id):
+                focused_at = window.last_focused_at
+                self.last_focused_window_with_progress_id = window.id
+        self.mark_tab_bar_dirty()
+        tm = self.tab_manager_ref()
+        if tm is not None:
+            tm.update_progress()
+
     def has_single_window_visible(self) -> bool:
         if self.current_layout.only_active_window_visible:
             return True
@@ -190,9 +212,14 @@ class Tab:  # {{{
     def take_over_from(self, other_tab: 'Tab') -> None:
         self.name, self.cwd = other_tab.name, other_tab.cwd
         self.enabled_layouts = list(other_tab.enabled_layouts)
-        if other_tab._current_layout_name:
-            self._set_current_layout(other_tab._current_layout_name)
         self._last_used_layout = other_tab._last_used_layout
+        if clname := other_tab._current_layout_name:
+            cl = other_tab.current_layout
+            other_tab._set_current_layout(clname)
+            cl.set_owner(self.os_window_id, self.id)
+            self.current_layout: Layout = cl
+            self._current_layout_name = clname
+            self.mark_tab_bar_dirty()
         for window in other_tab.windows:
             detach_window(other_tab.os_window_id, other_tab.id, window.id)
         self.windows = other_tab.windows
@@ -211,19 +238,34 @@ class Tab:  # {{{
         self.mark_tab_bar_dirty()
 
     def startup(self, session_tab: 'SessionTab') -> None:
+        target_tab = self
+        boss = get_boss()
         for window in session_tab.windows:
             spec = window.launch_spec
             if isinstance(spec, SpecialWindowInstance):
                 self.new_special_window(spec)
             else:
                 from .launch import launch
-                launch(get_boss(), spec.opts, spec.args, target_tab=self, force_target_tab=True)
+                launched_window = launch(boss, spec.opts, spec.args, target_tab=target_tab, force_target_tab=True)
             if window.resize_spec is not None:
                 self.resize_window(*window.resize_spec)
+            if window.focus_matching_window_spec:
+                for w in boss.match_windows(window.focus_matching_window_spec, launched_window or boss.active_window):
+                    tab = w.tabref()
+                    if tab:
+                        target_tab = tab or self
+                        tm = tab.tab_manager_ref()
+                        if tm and boss.active_tab is not target_tab:
+                            tm.set_active_tab(target_tab)
+                        if target_tab.active_window is not w:
+                            target_tab.set_active_window(w)
+                        if current_focused_os_window_id() != w.os_window_id:
+                            focus_os_window(w.os_window_id, True)
 
-        self.windows.set_active_window_group_for(self.windows.all_windows[session_tab.active_window_idx])
+        with suppress(IndexError):
+            self.windows.set_active_window_group_for(self.windows.all_windows[session_tab.active_window_idx])
 
-    def serialize_state(self) -> Dict[str, Any]:
+    def serialize_state(self) -> dict[str, Any]:
         return {
             'version': 1,
             'id': self.id,
@@ -249,11 +291,11 @@ class Tab:  # {{{
             tm.mark_tab_bar_dirty()
 
     @property
-    def active_window(self) -> Optional[Window]:
+    def active_window(self) -> Window | None:
         return self.windows.active_window
 
     @property
-    def active_window_for_cwd(self) -> Optional[Window]:
+    def active_window_for_cwd(self) -> Window | None:
         return self.windows.active_group_main
 
     @property
@@ -265,19 +307,11 @@ class Tab:  # {{{
     def effective_title(self) -> str:
         return self.name or self.title
 
-    @property
-    def number_of_windows_with_running_programs(self) -> int:
-        ans = 0
-        for window in self:
-            if window.has_running_program:
-                ans += 1
-        return ans
-
-    def get_cwd_of_active_window(self, oldest: bool = False) -> Optional[str]:
+    def get_cwd_of_active_window(self, oldest: bool = False) -> str | None:
         w = self.active_window
         return w.get_cwd_of_child(oldest) if w else None
 
-    def get_exe_of_active_window(self, oldest: bool = False) -> Optional[str]:
+    def get_exe_of_active_window(self, oldest: bool = False) -> str | None:
         w = self.active_window
         return w.get_exe_of_child(oldest) if w else None
 
@@ -393,8 +427,8 @@ class Tab:  # {{{
         else:
             self.goto_layout(layout_name)
 
-    def resize_window_by(self, window_id: int, increment: float, is_horizontal: bool) -> Optional[str]:
-        increment_as_percent = self.current_layout.bias_increment_for_cell(self.windows, window_id, is_horizontal) * increment
+    def resize_window_by(self, window_id: int, increment: float, is_horizontal: bool) -> str | None:
+        increment_as_percent = self.current_layout.bias_increment_for_cell(self.windows, is_horizontal) * increment
         if self.current_layout.modify_size_of_window(self.windows, window_id, increment_as_percent, is_horizontal):
             self.relayout()
             return None
@@ -417,7 +451,7 @@ class Tab:  # {{{
         if w is not None and self.resize_window_by(
                 w.id, increment, is_horizontal) is not None:
             if get_options().enable_audio_bell:
-                ring_bell()
+                ring_bell(self.os_window_id)
 
     @ac('win', 'Reset window sizes undoing any dynamic resizing of windows')
     def reset_window_sizes(self) -> None:
@@ -429,21 +463,23 @@ class Tab:  # {{{
         ret = self.current_layout.layout_action(action_name, args, self.windows)
         if ret is None:
             if get_options().enable_audio_bell:
-                ring_bell()
+                ring_bell(self.os_window_id)
             return
         self.relayout()
 
     def launch_child(
         self,
         use_shell: bool = False,
-        cmd: Optional[List[str]] = None,
-        stdin: Optional[bytes] = None,
-        cwd_from: Optional[CwdRequest] = None,
-        cwd: Optional[str] = None,
-        env: Optional[Dict[str, str]] = None,
+        cmd: list[str] | None = None,
+        stdin: bytes | None = None,
+        cwd_from: CwdRequest | None = None,
+        cwd: str | None = None,
+        env: dict[str, str] | None = None,
         is_clone_launch: str = '',
         add_listen_on_env_var: bool = True,
         hold: bool = False,
+        pass_fds: tuple[int, ...] = (),
+        remote_control_fd: int = -1,
     ) -> Child:
         check_for_suitability = True
         if cmd is None:
@@ -484,46 +520,59 @@ class Tab:  # {{{
                                     cmd[:0] = [resolved_shell(get_options())[0]]
                                 cmd[0] = which(cmd[0]) or cmd[0]
                                 cmd = cmdline_for_hold(cmd)
-        fenv: Dict[str, str] = {}
+        fenv: dict[str, str] = {}
         if env:
             fenv.update(env)
         fenv['KITTY_WINDOW_ID'] = str(next_window_id())
         pwid = platform_window_id(self.os_window_id)
         if pwid is not None:
             fenv['WINDOWID'] = str(pwid)
-        ans = Child(cmd, cwd or self.cwd, stdin, fenv, cwd_from, is_clone_launch=is_clone_launch, add_listen_on_env_var=add_listen_on_env_var, hold=hold)
+        ans = Child(
+                cmd, cwd or self.cwd, stdin, fenv, cwd_from, is_clone_launch=is_clone_launch,
+                add_listen_on_env_var=add_listen_on_env_var, hold=hold, pass_fds=pass_fds, remote_control_fd=remote_control_fd)
         ans.fork()
         return ans
 
-    def _add_window(self, window: Window, location: Optional[str] = None, overlay_for: Optional[int] = None, overlay_behind: bool = False) -> None:
-        self.current_layout.add_window(self.windows, window, location, overlay_for, put_overlay_behind=overlay_behind)
+    def _add_window(
+        self, window: Window, location: str | None = None, overlay_for: int | None = None,
+        overlay_behind: bool = False, bias: float | None = None, next_to: Window | None = None,
+    ) -> None:
+        self.current_layout.add_window(self.windows, window, location, overlay_for, put_overlay_behind=overlay_behind, bias=bias, next_to=next_to)
+        if overlay_behind and (w := self.active_window):
+            set_redirect_keys_to_overlay(self.os_window_id, self.id, w.id, window.id)
+            buffer_keys_in_window(self.os_window_id, self.id, window.id, True)
+            window.keys_redirected_till_ready_from = w.id
         self.mark_tab_bar_dirty()
         self.relayout()
 
     def new_window(
         self,
         use_shell: bool = True,
-        cmd: Optional[List[str]] = None,
-        stdin: Optional[bytes] = None,
-        override_title: Optional[str] = None,
-        cwd_from: Optional[CwdRequest] = None,
-        cwd: Optional[str] = None,
-        overlay_for: Optional[int] = None,
-        env: Optional[Dict[str, str]] = None,
-        location: Optional[str] = None,
-        copy_colors_from: Optional[Window] = None,
+        cmd: list[str] | None = None,
+        stdin: bytes | None = None,
+        override_title: str | None = None,
+        cwd_from: CwdRequest | None = None,
+        cwd: str | None = None,
+        overlay_for: int | None = None,
+        env: dict[str, str] | None = None,
+        location: str | None = None,
+        copy_colors_from: Window | None = None,
         allow_remote_control: bool = False,
-        marker: Optional[str] = None,
-        watchers: Optional[Watchers] = None,
+        marker: str | None = None,
+        watchers: Watchers | None = None,
         overlay_behind: bool = False,
         is_clone_launch: str = '',
-        remote_control_passwords: Optional[Dict[str, Sequence[str]]] = None,
+        remote_control_passwords: dict[str, Sequence[str]] | None = None,
         hold: bool = False,
+        bias: float | None = None,
+        pass_fds: tuple[int, ...] = (),
+        remote_control_fd: int = -1,
+        next_to: Window | None = None,
     ) -> Window:
         child = self.launch_child(
             use_shell=use_shell, cmd=cmd, stdin=stdin, cwd_from=cwd_from, cwd=cwd, env=env,
             is_clone_launch=is_clone_launch, add_listen_on_env_var=False if allow_remote_control and remote_control_passwords else True,
-            hold=hold,
+            hold=hold, pass_fds=pass_fds, remote_control_fd=remote_control_fd,
         )
         window = Window(
             self, child, self.args, override_title=override_title,
@@ -532,7 +581,7 @@ class Tab:  # {{{
         )
         # Must add child before laying out so that resize_pty succeeds
         get_boss().add_child(window)
-        self._add_window(window, location=location, overlay_for=overlay_for, overlay_behind=overlay_behind)
+        self._add_window(window, location=location, overlay_for=overlay_for, overlay_behind=overlay_behind, bias=bias, next_to=next_to)
         if marker:
             try:
                 window.set_marker(marker)
@@ -544,9 +593,12 @@ class Tab:  # {{{
     def new_special_window(
             self,
             special_window: SpecialWindowInstance,
-            location: Optional[str] = None,
-            copy_colors_from: Optional[Window] = None,
+            location: str | None = None,
+            copy_colors_from: Window | None = None,
             allow_remote_control: bool = False,
+            remote_control_passwords: dict[str, Sequence[str]] | None = None,
+            pass_fds: tuple[int, ...] = (),
+            remote_control_fd: int = -1,
     ) -> Window:
         return self.new_window(
             use_shell=False, cmd=special_window.cmd, stdin=special_window.stdin,
@@ -554,7 +606,7 @@ class Tab:  # {{{
             cwd_from=special_window.cwd_from, cwd=special_window.cwd, overlay_for=special_window.overlay_for,
             env=special_window.env, location=location, copy_colors_from=copy_colors_from,
             allow_remote_control=allow_remote_control, watchers=special_window.watchers, overlay_behind=special_window.overlay_behind,
-            hold=special_window.hold,
+            hold=special_window.hold, remote_control_passwords=remote_control_passwords, pass_fds=pass_fds, remote_control_fd=remote_control_fd,
         )
 
     @ac('win', 'Close all windows in the tab other than the currently active window')
@@ -568,8 +620,8 @@ class Tab:  # {{{
     def move_window_to_top_of_group(self, window: Window) -> bool:
         return self.windows.move_window_to_top_of_group(window)
 
-    def overlay_parent(self, window: Window) -> Optional[Window]:
-        prev: Optional[Window] = None
+    def overlay_parent(self, window: Window) -> Window | None:
+        prev: Window | None = None
         for x in self.windows.windows_in_group_of(window):
             if x is window:
                 break
@@ -588,7 +640,7 @@ class Tab:  # {{{
         if active_window:
             self.title_changed(active_window)
 
-    def detach_window(self, window: Window) -> Tuple[Window, ...]:
+    def detach_window(self, window: Window) -> tuple[Window, ...]:
         windows = list(self.windows.windows_in_group_of(window))
         windows.sort(key=attrgetter('id'))  # since ids increase in order of creation
         for w in reversed(windows):
@@ -600,10 +652,10 @@ class Tab:  # {{{
         attach_window(self.os_window_id, self.id, window.id)
         self._add_window(window)
 
-    def set_active_window(self, x: Union[Window, int], for_keep_focus: Optional[Window] = None) -> None:
+    def set_active_window(self, x: Window | int, for_keep_focus: Window | None = None) -> None:
         self.windows.set_active_window_group_for(x, for_keep_focus=for_keep_focus)
 
-    def get_nth_window(self, n: int) -> Optional[Window]:
+    def get_nth_window(self, n: int) -> Window | None:
         if self.windows:
             return self.current_layout.nth_window(self.windows, n)
         return None
@@ -680,7 +732,7 @@ class Tab:  # {{{
 
     prev_window = previous_window
 
-    def most_recent_group(self, groups: Sequence[int]) -> Optional[int]:
+    def most_recent_group(self, groups: Sequence[int]) -> int | None:
         groups_set = frozenset(groups)
 
         for window_id in reversed(self.windows.active_window_history):
@@ -698,7 +750,7 @@ class Tab:  # {{{
         ids = tuple(reversed(self.windows.active_window_history))
         return ids[min(n - 1, len(ids) - 1)] if ids else 0
 
-    def neighboring_group_id(self, which: EdgeLiteral) -> Optional[int]:
+    def neighboring_group_id(self, which: EdgeLiteral) -> int | None:
         neighbors = self.current_layout.neighbors(self.windows)
         candidates = neighbors.get(which)
         if candidates:
@@ -726,7 +778,7 @@ class Tab:  # {{{
             map ctrl+left move_window left
             map ctrl+down move_window bottom
         ''')
-    def move_window(self, delta: Union[EdgeLiteral, int] = 1) -> None:
+    def move_window(self, delta: EdgeLiteral | int = 1) -> None:
         if isinstance(delta, int):
             if self.current_layout.move_window(self.windows, delta):
                 self.relayout()
@@ -745,7 +797,7 @@ class Tab:  # {{{
                     self.relayout()
 
     @property
-    def all_window_ids_except_active_window(self) -> Set[int]:
+    def all_window_ids_except_active_window(self) -> set[int]:
         all_window_ids = {w.id for w in self}
         aw = self.active_window
         if aw is not None:
@@ -757,7 +809,7 @@ class Tab:  # {{{
         over the windows for easy selection in this mode. See :opt:`visual_window_select_characters`.
         ''')
     def focus_visible_window(self) -> None:
-        def callback(tab: Optional[Tab], window: Optional[Window]) -> None:
+        def callback(tab: Tab | None, window: Window | None) -> None:
             if tab and window:
                 tab.set_active_window(window)
 
@@ -765,7 +817,7 @@ class Tab:  # {{{
 
     @ac('win', 'Swap the current window with another window in the current tab, selected visually. See :opt:`visual_window_select_characters`')
     def swap_with_window(self) -> None:
-        def callback(tab: Optional[Tab], window: Optional[Window]) -> None:
+        def callback(tab: Tab | None, window: Window | None) -> None:
             if tab and window:
                 tab.swap_active_window_with(window.id)
         get_boss().visual_window_select_action(self, callback, 'Choose window to swap with', only_window_ids=self.all_window_ids_except_active_window)
@@ -784,7 +836,7 @@ class Tab:  # {{{
     def move_window_backward(self) -> None:
         self.move_window(-1)
 
-    def list_windows(self, self_window: Optional[Window] = None, window_filter: Optional[Callable[[Window], bool]] = None) -> Generator[WindowDict, None, None]:
+    def list_windows(self, self_window: Window | None = None, window_filter: Callable[[Window], bool] | None = None) -> Generator[WindowDict, None, None]:
         active_window = self.active_window
         for w in self:
             if window_filter is None or window_filter(w):
@@ -793,7 +845,7 @@ class Tab:  # {{{
                     is_focused=w.os_window_id == current_focused_os_window_id() and w is active_window,
                     is_self=w is self_window)
 
-    def list_groups(self) -> List[Dict[str, Any]]:
+    def list_groups(self) -> list[dict[str, Any]]:
         return [g.as_simple_dict() for g in self.windows.groups]
 
     def matches_query(self, field: str, query: str, active_tab_manager: Optional['TabManager'] = None) -> bool:
@@ -865,15 +917,18 @@ class Tab:  # {{{
 class TabManager:  # {{{
 
     confirm_close_window_id: int = 0
+    num_of_windows_with_progress: int = 0
+    total_progress: int = 0
+    has_indeterminate_progress: bool = False
 
-    def __init__(self, os_window_id: int, args: CLIOptions, wm_class: str, wm_name: str, startup_session: Optional[SessionType] = None):
+    def __init__(self, os_window_id: int, args: CLIOptions, wm_class: str, wm_name: str, startup_session: SessionType | None = None):
         self.os_window_id = os_window_id
         self.wm_class = wm_class
         self.recent_mouse_events: Deque[TabMouseEvent] = deque()
         self.wm_name = wm_name
         self.args = args
         self.tab_bar_hidden = get_options().tab_bar_style == 'hidden'
-        self.tabs: List[Tab] = []
+        self.tabs: list[Tab] = []
         self.active_tab_history: Deque[int] = deque()
         self.tab_bar = TabBar(self.os_window_id)
         self._active_tab_idx = 0
@@ -893,7 +948,7 @@ class TabManager:  # {{{
         if new_active_tab_idx == self._active_tab_idx:
             return
         try:
-            old_active_tab: Optional[Tab] = self.tabs[self._active_tab_idx]
+            old_active_tab: Tab | None = self.tabs[self._active_tab_idx]
         except Exception:
             old_active_tab = None
         else:
@@ -901,7 +956,7 @@ class TabManager:  # {{{
             add_active_id_to_history(self.active_tab_history, old_active_tab.id)
         self._active_tab_idx = new_active_tab_idx
         try:
-            new_active_tab: Optional[Tab] = self.tabs[self._active_tab_idx]
+            new_active_tab: Tab | None = self.tabs[self._active_tab_idx]
         except Exception:
             new_active_tab = None
         if old_active_tab is not new_active_tab:
@@ -974,7 +1029,7 @@ class TabManager:  # {{{
             tab.relayout_borders()
         self.mark_tab_bar_dirty()
 
-    def set_active_tab(self, tab: Tab, for_keep_focus: Optional[Tab] = None) -> bool:
+    def set_active_tab(self, tab: Tab, for_keep_focus: Tab | None = None) -> bool:
         try:
             idx = self.tabs.index(tab)
         except Exception:
@@ -1003,7 +1058,7 @@ class TabManager:  # {{{
                     self.set_active_tab(x)
                     break
 
-    def tab_at_location(self, loc: str) -> Optional[Tab]:
+    def tab_at_location(self, loc: str) -> Tab | None:
         if loc == 'prev':
             if self.active_tab_history:
                 old_active_tab_id = self.active_tab_history[-1]
@@ -1021,17 +1076,17 @@ class TabManager:  # {{{
             tab_num = max(0, len(self.tabs) - 1)
         if tab_num >= 0:
             self.set_active_tab_idx(tab_num)
-        else:
+        elif self.active_tab_history:
             try:
                 old_active_tab_id = self.active_tab_history[tab_num]
             except IndexError:
-                return
+                old_active_tab_id = self.active_tab_history[0]
             for idx, tab in enumerate(self.tabs):
                 if tab.id == old_active_tab_id:
                     self.set_active_tab_idx(idx)
                     break
 
-    def nth_active_tab(self, n: int = 0) -> Optional[Tab]:
+    def nth_active_tab(self, n: int = 0) -> Tab | None:
         if n <= 0:
             return self.active_tab
         tab_ids = tuple(reversed(self.active_tab_history))
@@ -1044,9 +1099,9 @@ class TabManager:  # {{{
         return len(self.tabs)
 
     def list_tabs(
-        self, self_window: Optional[Window] = None,
-        tab_filter: Optional[Callable[[Tab], bool]] = None,
-        window_filter: Optional[Callable[[Window], bool]] = None
+        self, self_window: Window | None = None,
+        tab_filter: Callable[[Tab], bool] | None = None,
+        window_filter: Callable[[Window], bool] | None = None
     ) -> Generator[TabDict, None, None]:
         active_tab = self.active_tab
         for tab in self:
@@ -1067,7 +1122,7 @@ class TabManager:  # {{{
                         'active_window_history': list(tab.windows.active_window_history),
                     }
 
-    def serialize_state(self) -> Dict[str, Any]:
+    def serialize_state(self) -> dict[str, Any]:
         return {
             'version': 1,
             'id': self.os_window_id,
@@ -1076,26 +1131,19 @@ class TabManager:  # {{{
         }
 
     @property
-    def active_tab(self) -> Optional[Tab]:
+    def active_tab(self) -> Tab | None:
         try:
             return self.tabs[self.active_tab_idx] if self.tabs else None
         except Exception:
             return None
 
     @property
-    def active_window(self) -> Optional[Window]:
+    def active_window(self) -> Window | None:
         t = self.active_tab
         if t is not None:
             return t.active_window
         return None
 
-    @property
-    def number_of_windows_with_running_programs(self) -> int:
-        count = 0
-        for tab in self:
-            count += tab.number_of_windows_with_running_programs
-        return count
-
     @property
     def number_of_windows(self) -> int:
         count = 0
@@ -1103,7 +1151,7 @@ class TabManager:  # {{{
             count += len(tab)
         return count
 
-    def tab_for_id(self, tab_id: int) -> Optional[Tab]:
+    def tab_for_id(self, tab_id: int) -> Tab | None:
         for t in self.tabs:
             if t.id == tab_id:
                 return t
@@ -1122,8 +1170,8 @@ class TabManager:  # {{{
 
     def new_tab(
         self,
-        special_window: Optional[SpecialWindowInstance] = None,
-        cwd_from: Optional[CwdRequest] = None,
+        special_window: SpecialWindowInstance | None = None,
+        cwd_from: CwdRequest | None = None,
         as_neighbor: bool = False,
         empty_tab: bool = False,
         location: str = 'last'
@@ -1196,7 +1244,7 @@ class TabManager:  # {{{
         tab.destroy()
 
     @property
-    def tab_bar_data(self) -> List[TabBarData]:
+    def tab_bar_data(self) -> list[TabBarData]:
         at = self.active_tab
         ans = []
         for t in self.tabs:
@@ -1212,7 +1260,8 @@ class TabManager:  # {{{
                 title, t is at, needs_attention, t.id,
                 len(t), t.num_window_groups, t.current_layout.name or '',
                 has_activity_since_last_focus, t.active_fg, t.active_bg,
-                t.inactive_fg, t.inactive_bg
+                t.inactive_fg, t.inactive_bg, t.num_of_windows_with_progress,
+                t.total_progress, t.last_focused_window_with_progress_id,
             ))
         return ans
 
@@ -1244,8 +1293,20 @@ class TabManager:  # {{{
         if len(self.recent_mouse_events) > 5:
             self.recent_mouse_events.popleft()
 
+    def update_progress(self) -> None:
+        self.num_of_windows_with_progress = 0
+        self.total_progress = 0
+        self.has_indeterminate_progress = False
+        for tab in self:
+            if tab.num_of_windows_with_progress:
+                self.total_progress += tab.total_progress
+                self.num_of_windows_with_progress += tab.num_of_windows_with_progress
+            if tab.has_indeterminate_progress:
+                self.has_indeterminate_progress = True
+        get_boss().update_progress_in_dock()
+
     @property
-    def tab_bar_rects(self) -> Tuple[Border, ...]:
+    def tab_bar_rects(self) -> tuple[Border, ...]:
         return self.tab_bar.blank_rects if self.tab_bar_should_be_visible else ()
 
     def destroy(self) -> None:
